/*
 * (C) Copyright IBM Corp. 2003    All Rights Reserved.
 * simfail.c ($Revision: 1.4 $)
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corp.
 *
 * Author(s): S. Bader <shbader@de.ibm.com>
 *
 * This file is part of MDMPD.
 *
 * MDMPD is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MDMPD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MDMPD; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/kdev_t.h>
#include <linux/raid/md_u.h>

#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <mdstat.h>
#include <mdctrl.h>

#define ERR(t, ...)	fprintf(stderr, t, ## __VA_ARGS__)
#if DEBUG > 4
#define DBG(t, ...)	fprintf(stderr, "%s(%d) in %s: " t, \
				__FILE__ , __LINE__ , __FUNCTION__ , \
				## __VA_ARGS__)
#else
#define DBG(t, ...)
#endif

/* License statement */
static char *license = "\n(C) Copyright IBM Corp. 2003    All Rights Reserved.\nmdmpd.c ($Revision: 1.5 $)\nUS Government Users Resticted Rights -\nUse, Duplication or disclosure restricted by\nGSA ADP Schedule Contract with IBM Corp\n";

struct mds_info		mdi;
int			shutdown;

int md_fail_disk(int fd, mdu_disk_info_t *disk) {
	DBG("Disk%d set to faulty\n", disk->number);
	return ioctl(
		fd,
		SET_DISK_FAULTY,
		MKDEV(disk->major, disk->minor)
	);
}

int simulate_failure(struct mds_array *mda) {
	static mdu_disk_info_t	disk;
	static char		path[80];
	int			fd;
	int			rc;
	int			i;

	sprintf(path, "/var/run/mdmpd/md%d", mda->nr);
	DBG("Open %s\n", path);
	if((fd = open(path, O_RDONLY)) < 0)
		return -1;

	for(i = 0; i < MAX_MD_DISKS; i++) {
		if(mda->disk[i].status == disk_missing)
			continue;
		disk.number = i;
		if(ioctl(fd, GET_DISK_INFO, &disk) != 0)
			continue;
		DBG("Disk%d: state=%d\n", i, disk.state);
		if((disk.state & 6) != 6)
			continue;

		rc = md_fail_disk(fd, &disk);
		break;
	}
	close(fd);

	return rc;
}

void sighandler(int signal) {
	fprintf(stderr, "Catched signal %d\n", signal);
	shutdown = 1;
}

int main(void) {
	int			rc;
	int			i;
	struct sigaction	sa;

	(void) license;

	if(mdstat_register(MDSTAT_TYPE_WITHOUT_EVENTS) < 0)
		return -1;

	sa.sa_handler = sighandler;
	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);

	memset(&mdi, 0, sizeof(mdi));
	mlockall(MCL_CURRENT);

	shutdown = 0;
	while(!shutdown) {
		sleep(5);
		if(shutdown)
			break;

		rc = mdstat_get(&mdi);

		for(i = 0; i < MAX_MD_ARRAYS; i++) {
			if(mdi.md[i].level != multipath)
				continue;

			simulate_failure(&mdi.md[i]);
		}
	}

	munlockall();

	mdstat_unregister();

	return rc;
}
